#include "sxBoot.h"
#include "sxPhysMem.h"
#include "sxConsole.h"
#include "sxMain.h"

#define SX_DUMP_NUMPERLINE	7
#define SX_PHYS_1MB_MARK	((1024 * 1024) / SX_PHYS_PAGE_SIZE) / 32
#define SX_PHYS_16MB_MARK	((1024 * 1024 * 16) / SX_PHYS_PAGE_SIZE) / 32
#define SX_PHYS_MAPSIZE		(SX_PHYS_MAX_PAGES / (sizeof(sx_uint32) * 8))

#define SX_SCALE_TO_INDEX(addr, idx, bit)					\
	bit = ((sx_uint32)addr / SX_PHYS_PAGE_SIZE) % 32;		\
	idx = (((sx_uint32)addr / SX_PHYS_PAGE_SIZE) - bit) / 32;

#define SX_SCALE_FROM_INDEX(idx, bit)						\
	((((idx) * 32) * SX_PHYS_PAGE_SIZE) + (bit * SX_PHYS_PAGE_SIZE))

#define SX_PHYS_CHECK_ALIGN(x) \
	(SX_PHYS_ALIGN(x, SX_PHYS_PAGE_ALIGN) == (sx_uint32)x)

#define SX_PHYS_WARN_MISALIGN \
	SxConPrint("WARNING: physical address misaligned!\n");

/* on a x86, this table requires exaclty 128 KB of memory
 * (in the .bss section), regardless of wether 4K or 4M
 * pages are used - in both cases there is the same number
 * of pages phsically addressable. */

static sx_uint32 sxMemoryMap[SX_PHYS_MAPSIZE];
static sx_uint32 sxAvailPages;
static sx_uint32 sxUsedPages;

void SxPhysMemInit() {

	/* initially, all memory is allocated. One needs to add
	 * free memory regions with SxPhysMemAddRegion, for example
	 * taken from the GRUM mmap (type 1 entries) */

	register sx_uint32* ptr = sxMemoryMap;
	register sx_uint32 l = SX_PHYS_MAPSIZE;

	while(--l) {
		*ptr++ = 0xFFFFFFFF;
	}
}

void SxPhysMemAddRegion(void* base, sx_uint32 size) {
	sx_uint8* pages = (sx_uint8*)SX_PHYS_ALIGN(base, SX_PHYS_PAGE_ALIGN);
	size -= ((sx_uint32)pages - (sx_uint32)base);

	for(; size >= SX_PHYS_PAGE_SIZE; size -= SX_PHYS_PAGE_SIZE, pages += SX_PHYS_PAGE_SIZE) {
		sx_uint32 idx, bit;

		SX_SCALE_TO_INDEX(pages, idx, bit);
		SX_BITCLEAR(sxMemoryMap[idx], bit);

		++sxAvailPages;
	}
}

void* SxPhysMemPlacementAlloc(void* addr, sx_uint32 pages) {
	register sx_uint32 idx, bit;

	if(!SX_PHYS_CHECK_ALIGN(addr)) { SX_PHYS_WARN_MISALIGN; return SX_NULL; }

	SX_SCALE_TO_INDEX(addr, idx, bit);

	while(pages--) {
		if(idx > SX_PHYS_MAPSIZE) {
			SxConPrint("Address out of range.\n");
			return SX_NULL;
		}

		if(SX_BITTEST(sxMemoryMap[idx], bit)) {
			SxConPrint("Placement address already in use!\n");
			return SX_NULL;
		}

		SX_BITSET(sxMemoryMap[idx], bit);
		++sxUsedPages;

		++bit; if(bit == 32) { bit = 0; idx++; }
	}

	return addr;
}

void SxPhysMemFree(void* addr, sx_uint32 pages) {
	register sx_uint32 idx, bit;

	if(!SX_PHYS_CHECK_ALIGN(addr)) { SX_PHYS_WARN_MISALIGN; return; }

	SX_SCALE_TO_INDEX(addr, idx, bit);

	while(pages--) {
		if(idx > SX_PHYS_MAPSIZE) {
			SxConPrint("Address out of range.\n");
			return;
		}

		if(!SX_BITTEST(sxMemoryMap[idx], bit)) {
			SxConPrint("Invalid Free, address not in use!\n");
			return;
		}

		SX_BITCLEAR(sxMemoryMap[idx], bit);
		--sxUsedPages;

		++bit; if(bit == 32) { bit = 0; idx++; }
	}
}

void* SxPhysMemAlloc(sx_uint32 pages, sx_physmem_flag_t flags) {
	register sx_uint32 idx, bit = 0, p = pages;

	/* WARNING: This assumes that the machine has more than
	 * 16 MB RAM! Otherwise allocation will always fail, except
	 * for when epmBlowSixteenMb or epmBelowOneMb is specified
	 * explicitly. I could also make the search restart with a
	 * lowered limit if i can't find free pages above 16 MB */

	if(flags & epmBelowOneMb) {
		idx = 0;
	} else if(flags & epmBelowSixteenMb) {
		idx = SX_PHYS_1MB_MARK;
	} else {
		idx = SX_PHYS_16MB_MARK;
	}

	while(idx < SX_PHYS_MAPSIZE) {

		/* small check wether there is at least one free page
		 * at this index, otherwise we would waste time looking
		 * at 32 unavailable pages ... */

		if(sxMemoryMap[idx] ^ 0xFFFFFFFF) {
			register sx_uint32 i = idx, b = 0;

			while(i < SX_PHYS_MAPSIZE && p > 0) {
				if(SX_BITTEST(sxMemoryMap[i], b)) {

					/* reset to current location */
					p = pages;
					++b; if(b == 32) { idx = i; break; }

					bit = b;
				} else {
					--p;
					++b; if(b == 32) { b = 0; i++; }
				}
			}

			if(!p) break;
		}

		idx++;

		if(flags & epmBelowOneMb && idx > SX_PHYS_1MB_MARK) break;
		if(flags & epmBelowSixteenMb && idx > SX_PHYS_16MB_MARK) break;
	}

	if(p) {
		if(idx >= SX_PHYS_MAPSIZE) {
			SxCorePanic("Out of physical memory! Cannot allocate %d contiguous pages!\n", pages);
		} else if(idx >= SX_PHYS_1MB_MARK && flags & epmBelowOneMb) {
			SxConPrint("There are no %d contiguous pages below 1MB\n", pages);
		} else if(idx >= SX_PHYS_16MB_MARK && flags & epmBelowSixteenMb) {
			SxConPrint("There are no %d contiguous pages below 16MB (and above 1MB)\n", pages);
		} else {
			SxConPrint("Unknown error allocating %d pages!\n", pages);
		}
	} else {

		/* phew... found pages. from idx:bit onwards are enough free pages
		 * so now allocate them (whithout any checks, they have to be done
		 * while checking, see above!) */

		void * start = (void*)SX_SCALE_FROM_INDEX(idx, bit);
		p = pages;

		while(p--) {
			SX_BITSET(sxMemoryMap[idx], bit);
			++sxUsedPages;

			++bit; if(bit == 32) { bit = 0; ++idx; }
		}

		return start;
	}

	return SX_NULL;
}

void SxPhysMemDumpState(sx_uint32 idx, sx_uint32 lines) {
	sx_uint32* current = &sxMemoryMap[idx];

	while(current < (&sxMemoryMap[idx] + (lines * SX_DUMP_NUMPERLINE))) {
		sx_uint8 i;

		for(i = 0; i < SX_DUMP_NUMPERLINE; ++i) {
			if(current > (sxMemoryMap + sizeof(sxMemoryMap))) {
				break;
			}

			SxConPrint(" %P", *current++);
		}
		SxConChar('\n');
	}
}

void SxPhysMemStat(sx_uint32* free, sx_uint32* used) {
	if(free) *free = sxAvailPages - sxUsedPages;
	if(used) *used = sxUsedPages;
}